@@ -32,14 +32,14 @@
 #include <stdlib.h>
 #include <pwd.h>
 #include <errno.h>
-#include <libsexy/sexy.h>
-#include <libgnomevfs/gnome-vfs-utils.h>
+#include <gconf/gconf-client.h>
 
 #include <glib/gi18n-lib.h>
 #include <gtk/gtk.h>
 #include <dbus/dbus-glib.h>
 #include <dbus/dbus-glib-lowlevel.h>
 
+#include "sexy-url-label.h"
 #include "polkit-gnome-auth-dialog.h"
 
 struct _PolkitGnomeAuthDialogPrivate
@@ -335,6 +335,63 @@ polkit_gnome_auth_dialog_set_message (PolkitGnomeAuthDialog *auth_dialog, const
 	g_free (str);
 }
 
+static void
+retain_checkbox_set_defaults (PolkitGnomeAuthDialog *auth_dialog, const char *action_id)
+{
+	gboolean retain_authorization;
+        GConfClient *client;
+	GError *error;
+	GSList *action_list, *l;
+
+        client = gconf_client_get_default ();
+	retain_authorization = TRUE;
+
+	if (client == NULL) {
+		g_warning ("Error getting GConfClient");
+		goto out;
+	}
+
+	error = NULL;
+	retain_authorization = gconf_client_get_bool (client, KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION, &error);
+	if (error != NULL) {
+		g_warning ("Error getting key %s: %s",
+			   KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION,
+			   error->message);
+		g_error_free (error);
+		goto out;
+	}
+
+	/* check the blacklist */
+	if (!retain_authorization)
+		goto out;
+
+	action_list = gconf_client_get_list (client,
+					     KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION_BLACKLIST,
+					     GCONF_VALUE_STRING,
+					     &error);
+	if (error != NULL) {
+		g_warning ("Error getting key %s: %s",
+			   KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION_BLACKLIST,
+			   error->message);
+		g_error_free (error);
+		goto out;
+	}
+
+	for (l = action_list; l != NULL; l = l->next) {
+		const char *str = l->data;
+		if (strcmp (str, action_id) == 0) {
+			retain_authorization = FALSE;
+			break;
+		}
+	}
+	g_slist_foreach (action_list, (GFunc) g_free, NULL);
+	g_slist_free (action_list);
+
+out:
+	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (auth_dialog->priv->keep_privilege_check_button),
+				      retain_authorization);
+}
+
 static void
 polkit_gnome_auth_dialog_set_action_id (PolkitGnomeAuthDialog *auth_dialog, const char *action_id)
 {
@@ -348,6 +405,8 @@ polkit_gnome_auth_dialog_set_action_id (PolkitGnomeAuthDialog *auth_dialog, cons
 	str = g_strdup_printf (_("Click to edit %s"), action_id);
 	gtk_widget_set_tooltip_markup (auth_dialog->priv->privilege_desc_label, str);
 	g_free (str);
+
+	retain_checkbox_set_defaults (auth_dialog, action_id);
 }
 
 static void
@@ -380,16 +439,18 @@ static void
 polkit_gnome_auth_dialog_set_icon_name (PolkitGnomeAuthDialog *auth_dialog, const char *icon_name)
 {
 	GdkPixbuf *pixbuf;
+	GdkPixbuf *copy_pixbuf;
 	GdkPixbuf *vendor_pixbuf;
 
 	pixbuf = NULL;
+	copy_pixbuf = NULL;
 	vendor_pixbuf = NULL;
 
 	vendor_pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
-					   icon_name,
-					   48,
-					   0,
-					   NULL);
+						  icon_name,
+						  48,
+						  0,
+						  NULL);
 	if (vendor_pixbuf == NULL)
 		goto out;
 
@@ -401,21 +462,27 @@ polkit_gnome_auth_dialog_set_icon_name (PolkitGnomeAuthDialog *auth_dialog, cons
 	if (pixbuf == NULL)
 		goto out;
 
+	/* need to copy the pixbuf since we're modifying it */
+	copy_pixbuf = gdk_pixbuf_copy (pixbuf);
+	if (copy_pixbuf == NULL)
+		goto out;
+
 	/* blend the vendor icon in the bottom right quarter */
 	gdk_pixbuf_composite (vendor_pixbuf,
-			      pixbuf,
+			      copy_pixbuf,
 			      24, 24, 24, 24,
 			      24, 24, 0.5, 0.5,
 			      GDK_INTERP_BILINEAR,
 			      255);
-	
+
 	gtk_image_set_from_pixbuf (GTK_IMAGE (auth_dialog->priv->icon),
-				   pixbuf);
+				   copy_pixbuf);
 
 out:
-
 	if (pixbuf != NULL)
 		g_object_unref (pixbuf);
+	if (copy_pixbuf != NULL)
+		g_object_unref (copy_pixbuf);
 	if (vendor_pixbuf != NULL)
 		g_object_unref (vendor_pixbuf);
 }
@@ -542,7 +609,7 @@ static void
 vendor_url_activated (SexyUrlLabel *url_label, char *url, gpointer user_data)
 {
         if (url != NULL) {
-                gnome_vfs_url_show (url);
+                gtk_show_uri (NULL, url, GDK_CURRENT_TIME, NULL);
         }
 }
 
