@@ -33,6 +33,7 @@
 #include <glib/gi18n-lib.h>
 #include <glib-object.h>
 #include <gdk/gdkx.h>
+#include <gconf/gconf-client.h>
 
 #define DBUS_API_SUBJECT_TO_CHANGE
 #include <dbus/dbus-glib.h>
@@ -47,6 +48,8 @@
 
 #include "polkit-gnome-auth-dialog.h"
 
+#define KEY_AUTH_DIALOG_GRAB_KEYBOARD "/desktop/gnome/policykit/auth_dialog_grab_keyboard"
+
 static void do_cancel_auth (void);
 
 
@@ -359,22 +362,13 @@ get_exe_for_pid (pid_t pid)
 {
         char *result;
         char buf[PATH_MAX];
-        ssize_t len;
-        char proc_name[32];
 
-        result = NULL;
-
-        snprintf (proc_name, sizeof (proc_name), "/proc/%d/exe", pid);
-        len = readlink (proc_name, buf, sizeof (buf) - 1);
-        if (len == -1) {
-                goto out;
+        if (polkit_sysdeps_get_exe_for_pid_with_helper (pid, buf, sizeof buf) < 0) {
+                result = g_strdup (_("(unknown"));
+        } else {
+                result =  g_strdup (buf);
         }
-        g_assert (len >= 0 && len < PATH_MAX - 1);
 
-        buf[len] = '\0';
-        result = g_strdup (buf);
-
-out:
         return result;
 }
 
@@ -525,6 +519,8 @@ typedef struct {
         char **admin_users;
 
         char *admin_user_selected;
+
+        gboolean got_keyboard_grab;
 } UserData;
 
 static void
@@ -580,26 +576,25 @@ conversation_type (PolKitGrant *polkit_grant, PolKitResult auth_type, void *user
 static void
 _do_grab (GtkWidget *widget, GdkEvent *event, gpointer user_data)
 {
-        GdkGrabStatus ret;
+        UserData *ud = user_data;
+        GConfClient *client;
 
-        ret = gdk_pointer_grab (widget->window, 
-                                TRUE,
-                                GDK_POINTER_MOTION_MASK|GDK_BUTTON_MOTION_MASK,
-                                NULL,
-                                NULL,
-                                GDK_CURRENT_TIME);
-        if (ret != GDK_GRAB_SUCCESS) {
-                g_warning ("Couldn't grab the pointer; ret = %d", ret);
+        ud->got_keyboard_grab = FALSE;
+
+        client = gconf_client_get_default ();
+        if (gconf_client_get_bool (client, KEY_AUTH_DIALOG_GRAB_KEYBOARD, NULL)) { /* NULL-GError */
+                GdkGrabStatus ret;
+
+                ret = gdk_keyboard_grab (widget->window, 
+                                         FALSE,
+                                         GDK_CURRENT_TIME); /* FIXME: ideally we need a real timestamp */
+                if (ret != GDK_GRAB_SUCCESS) {
+                        g_warning ("Couldn't grab the keyboard; ret = %d", ret);
+                } else {
+                        g_debug ("Grabbed keyboard");
+                        ud->got_keyboard_grab = TRUE;
+                }
         }
-
-        ret = gdk_keyboard_grab (widget->window, 
-                                FALSE,
-                                GDK_CURRENT_TIME);
-        if (ret != GDK_GRAB_SUCCESS) {
-                g_warning ("Couldn't grab the pointer; ret = %d", ret);
-        }
-
-        g_debug ("Grabbed pointer and keyboard");
 }
 
 /*--------------------------------------------------------------------------------------------------------------*/
@@ -938,6 +933,76 @@ remove_watch (PolKitGrant *polkit_auth, int watch_id)
 static PolKitGrant *grant = NULL;
 static UserData *ud = NULL;
 
+static void
+add_to_blacklist (UserData *ud, const char *action_id)
+{
+        GSList *action_list, *l;
+        GConfClient *client;
+        GError *error;
+        gboolean never_retain_authorization;
+
+        client = gconf_client_get_default ();
+	if (client == NULL) {
+		g_warning ("Error getting GConfClient");
+		goto out;
+	}
+
+	error = NULL;
+	never_retain_authorization = !gconf_client_get_bool (client, KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION, &error);
+	if (error != NULL) {
+		g_warning ("Error getting key %s: %s",
+			   KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION,
+			   error->message);
+		g_error_free (error);
+		goto out;
+	}
+
+        if (never_retain_authorization)
+                goto out;
+
+        error = NULL;
+        action_list = gconf_client_get_list (client,
+                                             KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION_BLACKLIST,
+                                             GCONF_VALUE_STRING,
+                                             &error);
+        if (error != NULL) {
+                g_warning ("Error getting key %s: %s",
+                           KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION_BLACKLIST,
+                           error->message);
+                g_error_free (error);
+                goto out;
+        }
+
+        for (l = action_list; l != NULL; l = l->next) {
+                const char *str = l->data;
+                if (strcmp (str, action_id) == 0) {
+                        /* already there */
+                        goto done;
+                }
+        }
+
+        action_list = g_slist_append (action_list, g_strdup (action_id));
+
+        if (!gconf_client_set_list (client,
+                                    KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION_BLACKLIST,
+                                    GCONF_VALUE_STRING,
+                                    action_list,
+                                    &error)) {
+                g_warning ("Error setting key %s: %s",
+                           KEY_AUTH_DIALOG_RETAIN_AUTHORIZATION_BLACKLIST,
+                           error->message);
+                g_error_free (error);
+                error = NULL;
+        }
+
+done:
+        g_slist_foreach (action_list, (GFunc) g_free, NULL);
+        g_slist_free (action_list);
+
+out:
+        ;
+}
+
 static gboolean
 do_polkit_auth (PolKitContext  *pk_context,
                 DBusConnection *system_bus_connection,
@@ -1086,14 +1151,25 @@ try_again:
                 }
         }
 
+        if (ud->gained_privilege) {
+                /* add to blacklist if the user unchecked the "remember authorization" check box */
+                if ((ud->remember_always &&
+                     !polkit_gnome_auth_dialog_get_remember_always (POLKIT_GNOME_AUTH_DIALOG (ud->dialog))) ||
+                    (ud->remember_session &&
+                     !polkit_gnome_auth_dialog_get_remember_session (POLKIT_GNOME_AUTH_DIALOG (ud->dialog)))) {
+                        add_to_blacklist (ud, action_id);
+                }
+        }
+
         ret = ud->gained_privilege;
 
 error:
 
         /* Ungrab keyboard and pointer */
-        gdk_keyboard_ungrab (GDK_CURRENT_TIME);
-        gdk_pointer_ungrab (GDK_CURRENT_TIME);
-        g_debug ("Ungrabbed pointer and keyboard");
+        if (ud->got_keyboard_grab) {
+                gdk_keyboard_ungrab (GDK_CURRENT_TIME);
+                g_debug ("Ungrabbed keyboard");
+        }
 
         if (ud->dialog != NULL) {
                 gtk_widget_destroy (ud->dialog);
